'''
See COPYRIGHT.md for copyright information.

This module is provided as a "loader" for all the validation modules, and can be specified
as a plugin to load all the validate subdirectory plugins.

'''
from arelle.Version import authorLabel, copyrightLabel

__pluginInfo__ = {
    'name': 'Import all validate modules',
    'version': '1.0',
    'description': "This is imports plugin modules in the validate subtree.",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    # imported plugins
    'import': ("module_subtree",)
}



from __future__ import annotations

import itertools
import os
from dataclasses import dataclass, field


CONFORMANCE_SUITE_PATH_PREFIX = 'tests/resources/conformance_suites'


@dataclass(frozen=True)
class ConformanceSuiteConfig:
    file: str
    info_url: str
    local_filepath: str
    name: str
    additional_downloads: dict[str, str] = field(default_factory=dict)
    additional_plugins_by_prefix: list[tuple[str, frozenset[str]]] = field(default_factory=list)
    approximate_relative_timing: dict[str, float] = field(default_factory=dict)  # by uri
    args: list[str] = field(default_factory=list)
    capture_warnings: bool = True
    expected_empty_testcases: frozenset[str] = frozenset()
    expected_failure_ids: frozenset[str] = frozenset()
    expected_model_errors: frozenset[str] = frozenset()
    extract_path: str | None = None
    membership_url: str | None = None
    plugins: frozenset[str] = frozenset()
    public_download_url: str | None = None
    shards: int = 1
    url_replace: str | None = None

    def __post_init__(self) -> None:
        redundant_plugins = [(prefix, overlap)
            for prefix, additional_plugins in self.additional_plugins_by_prefix
            for overlap in [self.plugins & additional_plugins]
            if overlap]
        assert not redundant_plugins, \
            f'Plugins specified both as default and additional: {redundant_plugins}'
        overlapping_prefixes = [(p1, p2)
            for (p1, _), (p2, _) in itertools.combinations(self.additional_plugins_by_prefix, 2)
            if p1.startswith(p2) or p2.startswith(p1)]
        assert not overlapping_prefixes, \
            f'Overlapping prefixes are not supported: {overlapping_prefixes}'
        assert not (self.shards == 1 and self.additional_plugins_by_prefix), \
            'Cannot specify additional_plugins_by_prefix with only one shard.'
        plugin_combinations = len({plugins for _, plugins in self.additional_plugins_by_prefix}) + 1
        assert plugin_combinations <= self.shards, \
            'Too few shards to accommodate the number of plugin combinations:' \
            f' combinations={plugin_combinations} shards={self.shards}'

    @property
    def prefixed_extract_filepath(self) -> str | None:
        if self.extract_path is None:
            return None
        return os.path.join(CONFORMANCE_SUITE_PATH_PREFIX, self.extract_path)

    @property
    def prefixed_local_filepath(self) -> str:
        return os.path.join(CONFORMANCE_SUITE_PATH_PREFIX, self.local_filepath)



from __future__ import annotations
import os.path
import pytest
import tempfile
from collections import Counter, defaultdict
from contextlib import nullcontext
from dataclasses import dataclass
from heapq import heapreplace
from lxml import etree
from pathlib import PurePath, PurePosixPath
from typing import Any, Callable, ContextManager, TYPE_CHECKING, cast
from unittest.mock import patch
from zipfile import ZipFile

from arelle import ModelDocument, PackageManager, PluginManager
from arelle.CntlrCmdLine import parseAndRun
from arelle.FileSource import archiveFilenameParts
from arelle.WebCache import WebCache

from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig

if TYPE_CHECKING:
    from _pytest.mark import ParameterSet


def get_document_id(doc: ModelDocument.ModelDocument) -> str:
    file_source = doc.modelXbrl.fileSource
    basefile = getattr(file_source, 'basefile', None)
    if basefile is None:
        file_source = next(iter(file_source.referencedFileSources.values()), None)
        if file_source is not None:
            basefile = file_source.basefile
    if basefile is not None:
        doc_id = PurePath(doc.filepath).relative_to(basefile).as_posix()
    else:
        # give up
        parts = archiveFilenameParts(doc.filepath)
        assert parts is not None
        _, doc_id = parts
    return doc_id


def get_test_data(
        args: list[str],
        expected_failure_ids: frozenset[str] = frozenset(),
        expected_empty_testcases: frozenset[str] = frozenset(),
        expected_model_errors: frozenset[str] = frozenset()) -> list[ParameterSet]:
    """
    Produces a list of Pytest Params that can be fed into a parameterized pytest function

    :param args: The args to be parsed by arelle in order to correctly produce the desired result set
    :param expected_failure_ids: The set of string test IDs that are expected to fail
    :param expected_empty_testcases: The set of paths of empty testcases, relative to the suite zip
    :param expected_model_errors: The set of error codes expected to be in the ModelXbrl errors
    :return: A list of PyTest Params that can be used to run a parameterized pytest function
    """
    cntlr = parseAndRun(args)  # type: ignore[no-untyped-call]
    try:
        results = []
        test_cases_with_no_variations = set()
        test_cases_with_unrecognized_type = {}
        model_document = cntlr.modelManager.modelXbrl.modelDocument
        test_cases: list[ModelDocument.ModelDocument] = []
        if model_document.type in (ModelDocument.Type.TESTCASE, ModelDocument.Type.REGISTRYTESTCASE):
            test_cases.append(model_document)
        elif model_document.type == ModelDocument.Type.TESTCASESINDEX:
            referenced_documents = model_document.referencesDocument.keys()
            child_document_types = {doc.type for doc in referenced_documents}
            assert len(child_document_types) == 1, f'Multiple child document types found in {model_document.uri}: {child_document_types}.'
            [child_document_type] = child_document_types
            # the formula function registry conformance suite points to a list of registries, so we need to search one level deeper.
            if child_document_type == ModelDocument.Type.REGISTRY:
                referenced_documents = [case for doc in referenced_documents for case in doc.referencesDocument.keys()]
            test_cases = sorted(referenced_documents, key=lambda doc: doc.uri)
        elif model_document.type == ModelDocument.Type.INSTANCE:
            test_id = get_document_id(model_document)
            model_errors = sorted(cntlr.modelManager.modelXbrl.errors)
            expected_model_errors_list = sorted(expected_model_errors)
            param = pytest.param(
                {
                    'status': 'pass' if model_errors == expected_model_errors_list else 'fail',
                    'expected': expected_model_errors_list,
                    'actual': model_errors
                },
                id=test_id,
                marks=[pytest.mark.xfail()] if test_id in expected_failure_ids else []
            )
            results.append(param)
        else:
            raise Exception('Unhandled model document type: {}'.format(model_document.type))
        for test_case in test_cases:
            test_case_file_id = get_document_id(test_case)
            if test_case.type not in (ModelDocument.Type.TESTCASE, ModelDocument.Type.REGISTRYTESTCASE):
                test_cases_with_unrecognized_type[test_case_file_id] = test_case.type
            if not getattr(test_case, "testcaseVariations", None):
                test_cases_with_no_variations.add(test_case_file_id)
            else:
                for mv in test_case.testcaseVariations:
                    test_id = f'{test_case_file_id}:{mv.id}'
                    param = pytest.param(
                        {
                            'status': mv.status,
                            'expected': mv.expected,
                            'actual': mv.actual
                        },
                        id=test_id,
                        marks=[pytest.mark.xfail()] if test_id in expected_failure_ids else []
                    )
                    results.append(param)
        if test_cases_with_unrecognized_type:
            raise Exception(f"Some test cases have an unrecognized document type: {sorted(test_cases_with_unrecognized_type.items())}.")
        unrecognized_expected_empty_testcases = expected_empty_testcases.difference(map(get_document_id, test_cases))
        if unrecognized_expected_empty_testcases:
            raise Exception(f"Some expected empty test cases weren't found: {sorted(unrecognized_expected_empty_testcases)}.")
        unexpected_empty_testcases = test_cases_with_no_variations - expected_empty_testcases
        if unexpected_empty_testcases:
            raise Exception(f"Some test cases don't have any variations: {sorted(unexpected_empty_testcases)}.")
        test_id_frequencies = Counter(cast(str, p.id) for p in results)
        nonunique_test_ids = {test_id: count for test_id, count in test_id_frequencies.items() if count > 1}
        if nonunique_test_ids:
            raise Exception(f'Some test IDs are not unique.  Frequencies of nonunique test IDs: {nonunique_test_ids}.')
        nonexistent_expected_failure_ids = expected_failure_ids - test_id_frequencies.keys()
        if nonexistent_expected_failure_ids:
            raise Exception(f"Some expected failure IDs don't match any test cases: {sorted(nonexistent_expected_failure_ids)}.")
        return results
    finally:
        cntlr.modelManager.close()
        PackageManager.close()  # type: ignore[no-untyped-call]
        PluginManager.close()  # type: ignore[no-untyped-call]


original_normalize_url_function = WebCache.normalizeUrl


def normalize_url_function(config: ConformanceSuiteConfig) -> Callable[[WebCache, str, str | None], str]:
    def normalize_url(self: WebCache, url: str, base: str | None = None) -> str:
        assert config.url_replace is not None
        if url.startswith(config.url_replace):
            return url.replace(config.url_replace, f'{config.prefixed_local_filepath}/')
        return cast(str, original_normalize_url_function(self, url, base))
    return normalize_url


def get_test_shards(config: ConformanceSuiteConfig) -> list[tuple[list[str], frozenset[str]]]:
    test_path_prefix = PurePosixPath(config.file).parent
    with ZipFile(config.prefixed_local_filepath) as zip_file:
        tree = etree.parse(zip_file.open(config.file))
        testcases_element = tree.getroot() if tree.getroot().tag == 'testcases' else tree.find('testcases')
        assert testcases_element is not None
        test_root = testcases_element.get('root', '')
        # replace backslashes with forward slashes, e.g. in
        # 616-definition-syntax/616-14-RXP-definition-link-validations\616-14-RXP-definition-link-validations-testcase.xml
        path_strs = sorted(str(test_path_prefix / test_root / cast(str, e.get('uri')).replace('\\', '/'))
            for e in testcases_element.findall('testcase'))
    assert path_strs

    @dataclass(frozen=True)
    class Path:
        path: str
        plugins: tuple[str, ...]
        runtime: float
    paths_by_plugins: dict[tuple[str, ...], list[Path]] = defaultdict(list)
    for path_str in path_strs:
        path_plugins: set[str] = set()
        for prefix, additional_plugins in config.additional_plugins_by_prefix:
            if path_str.startswith(prefix):
                path_plugins.update(additional_plugins)
        paths_by_plugins[tuple(path_plugins)].append(
            Path(path=path_str, plugins=tuple(path_plugins), runtime=config.approximate_relative_timing.get(path_str, 1)))
    paths_in_runtime_order: list[Path] = sorted((path for paths in paths_by_plugins.values() for path in paths),
        key=lambda path: path.runtime, reverse=True)
    runtime_by_plugins: dict[tuple[str, ...], float] = {plugins: sum(path.runtime for path in paths)
        for plugins, paths in paths_by_plugins.items()}
    total_runtime = sum(runtime_by_plugins.values())
    shards_by_plugins: dict[tuple[str, ...], list[tuple[float, list[str]]]] = {}
    remaining_shards = config.shards
    for i, (plugins, _) in enumerate(paths_by_plugins.items()):
        n_shards = (remaining_shards
            if i == len(paths_by_plugins) - 1
            else 1 + round(runtime_by_plugins[plugins] / total_runtime * (config.shards - len(paths_by_plugins))))
        remaining_shards -= n_shards
        shards_by_plugins[plugins] = [(0, []) for _ in range(n_shards)]
    assert remaining_shards == 0
    for path in paths_in_runtime_order:
        shards_for_plugins = shards_by_plugins[path.plugins]
        shard_runtime, shard = shards_for_plugins[0]
        shard.append(path.path)
        heapreplace(shards_for_plugins, (shard_runtime + path.runtime, shard))
    assert shards_by_plugins.keys() == {()} | {tuple(plugins) for _, plugins in config.additional_plugins_by_prefix}
    shards = [(paths, frozenset(plugins))
        for plugins, runtimes_paths in shards_by_plugins.items()
        for _, paths in runtimes_paths]
    assert sorted(path for paths, _ in shards for path in paths) == path_strs
    return shards


def get_conformance_suite_test_results(
        config: ConformanceSuiteConfig,
        shard: int | None,
        log_to_file: bool = False,
        offline: bool = False) -> list[ParameterSet]:
    file_path = os.path.join(config.prefixed_local_filepath, config.file)
    assert shard is None or config.shards != 1, \
        'Conformance suite configuration must specify shards if --shard is passed'
    use_shards = shard is not None
    testcase_file_cm: Callable[[], ContextManager[Any]] = \
        (lambda: tempfile.NamedTemporaryFile(dir='.', mode='wb', suffix='.xml')) if use_shards else nullcontext  # type: ignore[assignment]
    with testcase_file_cm() as testcase_file:
        if use_shards:
            assert shard is not None
            shards = get_test_shards(config)
            test_paths, additional_plugins = shards[shard]
            zip_path = config.prefixed_local_filepath
            all_test_paths = {path for test_paths, _ in shards for path in test_paths}
            unrecognized_expected_empty_testcases = config.expected_empty_testcases - all_test_paths
            assert not unrecognized_expected_empty_testcases, f'Unrecognized expected empty testcases: {unrecognized_expected_empty_testcases}'
            expected_empty_testcases = config.expected_empty_testcases.intersection(test_paths)
            unrecognized_expected_failure_ids = {id.rsplit(':', 1)[0] for id in config.expected_failure_ids} - all_test_paths
            assert not unrecognized_expected_failure_ids, f'Unrecognized expected failure IDs: {unrecognized_expected_failure_ids}'
            expected_failure_ids = frozenset(id for id in config.expected_failure_ids if id.rsplit(':', 1)[0] in test_paths)

            root = etree.Element('testcases')
            tree = etree.ElementTree(root)
            pathlib_zip_path = PurePath(zip_path)
            for test_path in test_paths:
                etree.SubElement(root, 'testcase', uri=str((pathlib_zip_path / test_path).as_posix()))
            tree.write(testcase_file, encoding='utf-8', pretty_print=True, xml_declaration=True)
            testcase_file.flush()
            filename = testcase_file.name
        else:
            additional_plugins = frozenset().union(*(plugins for _, plugins in config.additional_plugins_by_prefix))
            filename = file_path
            expected_empty_testcases = config.expected_empty_testcases
            expected_failure_ids = config.expected_failure_ids
        plugins = config.plugins | additional_plugins
        args = [
            '--file', filename,
            '--keepOpen',
            '--validate',
        ]
        if plugins:
            args.extend(['--plugins', '|'.join(sorted(plugins))])
        if config.capture_warnings:
            args.append('--testcaseResultsCaptureWarnings')
        if log_to_file:
            shard_str = f'-s{shard}' if use_shards else ''
            args.extend([
                '--csvTestReport', f'conf-{config.name}{shard_str}-report.csv',
                '--logFile', f'conf-{config.name}{shard_str}-log.txt',
            ])
        if offline:
            args.extend(['--internetConnectivity', 'offline'])
        context_manager: ContextManager[Any]
        if config.url_replace:
            context_manager = patch('arelle.WebCache.WebCache.normalizeUrl', normalize_url_function(config))
        else:
            context_manager = nullcontext()
        with context_manager:
            return get_test_data(
                args + config.args,
                expected_failure_ids=expected_failure_ids,
                expected_empty_testcases=expected_empty_testcases,
                expected_model_errors=config.expected_model_errors
            )



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import Toplevel, PhotoImage, messagebox, N, S, E, W
try:
    from tkinter.ttk import Frame, Button
except ImportError:
    from ttk import Frame, Button
import tkinter.filedialog
import os
import regex as re
from arelle.ModelValue import dateTime
from arelle import XmlUtil
from arelle.UiUtil import gridCell, gridCombobox, label, checkbox
from arelle.CntlrWinTooltip import ToolTip
from arelle.PluginManager import pluginClassMethods
from arelle.UrlUtil import isValidAbsolute
from arelle.ValidateXbrlCalcs import ValidateCalcsMode as CalcsMode

'''
caller checks accepted, if True, caller retrieves url
'''
def getOptions(mainWin):
    dialog = DialogRssWatch(mainWin, mainWin.modelManager.rssWatchOptions)
    if dialog.accepted:
        mainWin.config["rssWatchOptions"] = dialog.options
        mainWin.saveConfig()

rssFeeds = {
    "US SEC US-GAAP Filings": "http://www.sec.gov/Archives/edgar/usgaap.rss.xml",
    "US SEC Inline Filings": "https://www.sec.gov/Archives/edgar/xbrl-inline.rss.xml",
    # obsolete: "US SEC Voluntary Filings": "http://www.sec.gov/Archives/edgar/xbrlrss.xml",
    # obsolete: "US SEC Voluntary Risk/Return Filings": "http://www.sec.gov/Archives/edgar/xbrl-rr-vfp.rss.xml",
    "US SEC Mutual Fund Risk/Return Filings": "http://www.sec.gov/Archives/edgar/xbrl-rr.rss.xml",
    "US SEC All Filings": "http://www.sec.gov/Archives/edgar/xbrlrss.all.xml",
            }

emailPattern = re.compile(
      r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"  # dot-atom
      r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016-\177])*"' # quoted-string
      r')@(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+[A-Z]{2,6}\.?$', re.IGNORECASE)


class DialogRssWatch(Toplevel):
    def __init__(self, mainWin, options):
        self.mainWin = mainWin
        parent = mainWin.parent
        super(DialogRssWatch, self).__init__(parent)
        self.parent = parent
        self.options = options
        parentGeometry = re.match("(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", parent.geometry())
        dialogX = int(parentGeometry.group(3))
        dialogY = int(parentGeometry.group(4))
        self.accepted = False

        self.transient(self.parent)
        self.title(_("RSS Feed Processing Control"))

        frame = Frame(self)

        # checkbox entries
        row = 1 # row number (to allow future plugins
        label(frame, 1, row, "RSS Feed:")
        feedSources = sorted(rssFeeds.keys())
        self.cellFeed = gridCombobox(frame, 2, row, options.get("feedSource",""), values=feedSources)
        self.cellFeed.grid(pady=2)
        ToolTip(self.cellFeed, text=_("Select an RSS feed to process for item matching, formulas, and validations as selected below"), wraplength=240)
        row += 1
        label(frame, 1, row, "Match fact text:")
        self.cellMatchText = gridCell(frame, 2, row, options.get("matchTextExpr",""))
        ToolTip(self.cellMatchText, text=_("Enter a regular expression to be matched to the text of each filing instance fact item. "
                                           "Regular expressions may contain patterns to detect, such as ab.?c, for any single character between b and c, or ab.*c for any number of characters between b and c."), wraplength=240)
        row += 1
        label(frame, 1, row, "Formula file:")
        self.cellFormulaFile = gridCell(frame,2, row, options.get("formulaFileUri",""))
        ToolTip(self.cellFormulaFile, text=_("Select a formula linkbase to to evaluate each filing.  "
                                             "The formula linkbase may contain one or more assertions, the results of which is recorded in the log file.  "
                                             "If unsuccessful assertion alerts are selected and an e-mail address provided, the recipient will be notified of filings with assertions that do not pass."), wraplength=240)
        openFileImage = PhotoImage(file=os.path.join(mainWin.imagesDir, "toolbarOpenFile.gif"))
        chooseFormulaFileButton = Button(frame, image=openFileImage, width=12, command=self.chooseFormulaFile)
        chooseFormulaFileButton.grid(row=row, column=3, sticky=W)
        row += 1
        openDatabaseImage = PhotoImage(file=os.path.join(mainWin.imagesDir, "toolbarOpenDatabase.gif"))
        for pluginXbrlMethod in pluginClassMethods("DialogRssWatch.FileChoices"):
            pluginXbrlMethod(self, frame, row, options, mainWin, openFileImage, openDatabaseImage)
            row += 1
        label(frame, 1, row, "Log file:")
        self.cellLogFile = gridCell(frame,2, row, options.get("logFileUri",""))
        ToolTip(self.cellLogFile, text=_("Select a log file in which to save an activity log, including validation results, matched item text, and formula results.\n\n "
                                         "If file ends in .xml it is xml-formatted, otherwise it is text. "), wraplength=240)
        chooseLogFileButton = Button(frame, image=openFileImage, width=12, command=self.chooseLogFile)
        chooseLogFileButton.grid(row=row, column=3, sticky=W)
        row += 1
        label(frame, 1, row, "E-mail alerts to:")
        self.cellEmailAddress = gridCell(frame,2, row, options.get("emailAddress",""))
        ToolTip(self.cellEmailAddress, text=_("Specify e-mail recipient(s) for alerts per below."), wraplength=240)
        propertiesImage = PhotoImage(file=os.path.join(mainWin.imagesDir, "toolbarProperties.gif"))
        smtpSetupButton = Button(frame, image=propertiesImage, width=12, command=self.setupSmtp)
        smtpSetupButton.grid(row=row, column=3, sticky=W)
        ToolTip(smtpSetupButton, text=_("Enter/edit settings of outgoing e-mail server (SMTP)."), wraplength=240)
        row += 1
        label(frame, 1, row, "Latest pub date:")
        pubdate = getattr(options,"latestPubDate",None)
        self.cellLatestPubDate = gridCell(frame,2, row, str(pubdate) if pubdate else "")
        ToolTip(self.cellLatestPubDate, text=_("Specify pub dateTime of last processed submission.  Next item to examine will be after this dateTime."), wraplength=240)
        clearImage = PhotoImage(file=os.path.join(mainWin.imagesDir, "toolbarDelete.gif"))
        clearPubDateButton = Button(frame, image=clearImage, width=12, command=self.clearPubDate)
        clearPubDateButton.grid(row=row, column=3, sticky=W)
        ToolTip(clearPubDateButton, text=_("Clear pub dateTime so that next cycle processes all entries in RSS feed."), wraplength=240)
        row += 1
        label(frame, 2, row, "Validate:")
        row += 1
        self.checkboxes = (
           checkbox(frame, 2, row,
                    "XBRL 2.1 and Dimensions rules",
                    "validateXbrlRules"),
           checkbox(frame, 2, row+1,
                    "Selected disclosure system rules",
                    "validateDisclosureSystemRules")
        )
        row += 2
        self.calcModeMenu = CalcsMode.menu() # dynamic, may change after initialization if language changes
        vals = list(self.calcModeMenu.keys())
        keys = dict((v,k) for k,v in self.calcModeMenu.items())
        self.validateCalcs = gridCombobox(frame, 2, row, keys.get(options.get("validateCalcs",0)), values=vals)
        self.validateCalcs.grid(pady=2,padx=(44,0))
        ToolTip(self.validateCalcs, text=_("Select calculations validation"), wraplength=240)
        row += 1
        self.checkboxes += (
           checkbox(frame, 2, row,
                    "Formula assertions",
                    "validateFormulaAssertions"),
           # Note: if adding to this list keep ModelFormulaObject.FormulaOptions in sync
        )
        row += 1
        for pluginXbrlMethod in pluginClassMethods("DialogRssWatch.ValidateChoices"):
            pluginXbrlMethod(self, frame, row, options, mainWin)
            row += 1
        label(frame, 2, row, "Alert on:")
        row += 1
        self.checkboxes += (
           checkbox(frame, 2, row,
                    "Facts with matching text",
                    "alertMatchedFactText"),
           checkbox(frame, 2, row+1,
                    "Unsuccessful formula assertions",
                    "alertAssertionUnsuccessful"),
           checkbox(frame, 2, row+2,
                    "Validation errors",
                    "alertValiditionError"),
           # Note: if adding to this list keep ModelFormulaObject.FormulaOptions in sync
           )
        row += 3

        mainWin.showStatus(None)

        cancelButton = Button(frame, text=_("Cancel"), width=8, command=self.close)
        ToolTip(cancelButton, text=_("Cancel operation, discarding changes and entries"))
        okButton = Button(frame, text=_("OK"), width=8, command=self.ok)
        ToolTip(okButton, text=_("Accept the options as entered above"))
        cancelButton.grid(row=row, column=1, columnspan=3, sticky=E, pady=3, padx=3)
        okButton.grid(row=row, column=1, columnspan=3, sticky=E, pady=3, padx=86)

        frame.grid(row=0, column=0, sticky=(N,S,E,W))
        frame.columnconfigure(2, weight=1)
        window = self.winfo_toplevel()
        window.columnconfigure(0, weight=1)
        self.geometry("+{0}+{1}".format(dialogX+50,dialogY+100))

        #self.bind("<Return>", self.ok)
        #self.bind("<Escape>", self.close)

        self.protocol("WM_DELETE_WINDOW", self.close)
        self.grab_set()
        self.wait_window(self)

    def chooseFormulaFile(self):
        filename = tkinter.filedialog.askopenfilename(
                            title=_("Choose formula file for RSS Watch"),
                            initialdir=self.options.get("rssWatchFormulaFileDir","."),
                            filetypes=[] if self.mainWin.isMac else [(_("XBRL files"), "*.*")],
                            defaultextension=".xml",
                            parent=self.parent)
        if filename:
            self.options["rssWatchFormulaFileDir"] = os.path.dirname(filename)
            self.cellFormulaFile.setValue(filename)

    def chooseLogFile(self):
        filename = tkinter.filedialog.asksaveasfilename(
                            title=_("Choose log file for RSS Watch"),
                            initialdir=self.options.get("rssWatchLogFileDir","."),
                            filetypes=[] if self.mainWin.isMac else [(_("Log files"), "*.*")],
                            defaultextension=".txt",
                            parent=self.parent)
        if filename:
            self.options["rssWatchLogFileDir"] = os.path.dirname(filename)
            self.cellLogFile.setValue(filename)

    def setupSmtp(self):
        from arelle.DialogUserPassword import askSmtp
        smtpSettings = askSmtp(self, self.options.get("smtpEmailSettings",()))
        if smtpSettings:
            self.options["smtpEmailSettings"] = smtpSettings

    def clearPubDate(self):
        self.cellLatestPubDate.setValue("")

    def checkEntries(self):
        errors = []
        if not self.cellFeed.value in rssFeeds and not isValidAbsolute(self.cellFeed.value):
            errors.append(_("RSS feed field contents invalid"))
        try:
            if self.cellMatchText.value:
                re.compile(self.cellMatchText.value)
        except Exception as err:
            errors.append(_("Match text field contents error {0}").format(err))
        if self.cellFormulaFile.value and not os.path.exists(self.cellFormulaFile.value):
            errors.append(_("Formula file not found {0}").format(self.cellFormulaFile.value))
        if self.cellLogFile.value and not os.path.exists(os.path.dirname(self.cellLogFile.value)):
            errors.append(_("Log file directory not found {0}").format(self.cellLogFile.value))
        if self.cellEmailAddress.value:
            if not emailPattern.match(self.cellEmailAddress.value):
                errors.append(_("E-mail address format error").format(self.cellLogFile.value))
            if not self.options.get("smtpEmailSettings"):
                errors.append(_("Please enter E-mail server settings (SMTP)"))
        if self.cellLatestPubDate.value and dateTime(self.cellLatestPubDate.value) is None:
            errors.append(_("Latest pub date field contents invalid"))
        if errors:
            messagebox.showwarning(_("Dialog validation error(s)"),
                                "\n ".join(errors), parent=self)
            return False
        return True

    def setOptions(self):
        # set formula options
        self.options["feedSource"] = self.cellFeed.value
        if self.cellFeed.value in rssFeeds:
            self.options["feedSourceUri"] = rssFeeds[self.cellFeed.value]
        else:
            self.options["feedSourceUri"] = self.cellFeed.value
        self.options["matchTextExpr"] = self.cellMatchText.value
        self.options["formulaFileUri"] = self.cellFormulaFile.value
        self.options["logFileUri"] = self.cellLogFile.value
        self.options["emailAddress"] = self.cellEmailAddress.value
        if self.cellLatestPubDate.value:
            # need datetime.datetime base class for pickling, not ModelValue class (unpicklable)
            self.options["latestPubDate"] = XmlUtil.datetimeValue(self.cellLatestPubDate.value)
        else:
            self.options["latestPubDate"] = None
        for checkbox in self.checkboxes:
            self.options[checkbox.attr] = checkbox.value
        self.options["validateCalcs"] = self.calcModeMenu[self.validateCalcs.value]

    def ok(self, event=None):
        if not self.checkEntries():
            return
        self.setOptions()
        self.accepted = True
        self.close()

    def close(self, event=None):
        self.parent.focus_set()
        self.destroy()



'''
Update Table Linkbase is an example of a plug-in to both GUI menu and command line/web service
that updates a table linkbase from Eurofiling 2010 syntax to XII 2011 PWD syntax and saves it.

See COPYRIGHT.md for copyright information.
'''
from arelle.Version import authorLabel, copyrightLabel

def generateUpdatedTableLB(dts, updatedTableLinkbaseFile):
    import os, io
    from arelle import XmlUtil, XbrlConst
    from arelle.ViewUtil import viewReferences, referenceURI
    from arelle.ModelRenderingObject import ModelEuAxisCoord

    if dts.fileSource.isArchive:
        dts.error("genTblLB:outFileIsArchive",
                 _("Updated Table Linkbase file cannot be an archive: %(tableLBOutputFile)s."),
                 modelObject=dts, tableLBOutputFile=updatedTableLinkbaseFile)
        return
    tblAxisRelSet = dts.relationshipSet(XbrlConst.euTableAxis)
    axisMbrRelSet = dts.relationshipSet(XbrlConst.euAxisMember)
    if len(tblAxisRelSet.modelRelationships) == 0:
        dts.error("genTblLB:noInputTables",
                 _("DTS does not contain Eurofiling 2010 tables and axes: %(entryFile)s."),
                 modelObject=dts, entryFile=dts.uri)
        return

    file = io.StringIO('''
<nsmap>
<link:linkbase
xmlns:label="http://xbrl.org/2008/label"
xmlns:gen="http://xbrl.org/2008/generic"
xmlns:df="http://xbrl.org/2008/filter/dimension"
xmlns:xlink="http://www.w3.org/1999/xlink"
xmlns:reference="http://xbrl.org/2008/reference"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xmlns:link="http://www.xbrl.org/2003/linkbase"
xmlns:table="http://xbrl.org/2011/table"
xmlns:formula="http://xbrl.org/2008/formula"
xsi:schemaLocation="
http://www.xbrl.org/2003/linkbase http://www.xbrl.org/2003/xbrl-linkbase-2003-12-31.xsd
http://xbrl.org/2008/generic http://www.xbrl.org/2008/generic-link.xsd
http://xbrl.org/2008/reference http://www.xbrl.org/2008/generic-reference.xsd
http://xbrl.org/2008/label http://www.xbrl.org/2008/generic-label.xsd
http://xbrl.org/2011/table http://www.xbrl.org/2011/table.xsd
http://xbrl.org/2008/filter/dimension http://www.xbrl.org/2008/dimension-filter.xsd">
  <link:arcroleRef arcroleURI="http://xbrl.org/arcrole/2011/table-filter" xlink:type="simple"
   xlink:href="http://www.xbrl.org/2011/table.xsd#table-filter"/>
  <link:arcroleRef arcroleURI="http://xbrl.org/arcrole/2011/table-axis" xlink:type="simple"
   xlink:href="http://www.xbrl.org/2011/table.xsd#table-axis"/>
  <link:arcroleRef arcroleURI="http://xbrl.org/arcrole/2011/axis-subtree" xlink:type="simple"
   xlink:href="http://www.xbrl.org/2011/table.xsd#axis-subtree"/>
  <link:arcroleRef arcroleURI="http://xbrl.org/arcrole/2011/axis-filter" xlink:type="simple"
   xlink:href="http://www.xbrl.org/2011/filter-axis.xsd#axis-filter"/>
</link:linkbase>
</nsmap>
<!--  Generated by Arelle(r) http://arelle.org -->
'''
     )
    from arelle.ModelObjectFactory import parser
    parser, parserLookupName, parserLookupClass = parser(dts,None)
    from lxml import etree
    xmlDocument = etree.parse(file,parser=parser,base_url=updatedTableLinkbaseFile)
    file.close()
    nsmapElt = xmlDocument.getroot()
    #xmlDocument.getroot().init(self)  ## is this needed ??
    for lbElement in  xmlDocument.iter(tag="{http://www.xbrl.org/2003/linkbase}linkbase"):
        break

    class DocObj:  # fake ModelDocument for namespaces
        def __init__(self):
            self.xmlRootElement = lbElement
            self.xmlDocument = xmlDocument
    docObj = DocObj()

    numELRs = 0
    numTables = 0

    def copyAttrs(fromElt, toElt, attrTags):
        for attr in attrTags:
            if fromElt.get(attr):
                toElt.set(attr, fromElt.get(attr))

    def generateTable(newLinkElt, newTblElt, srcTblElt, tblAxisRelSet, axisMbrRelSet, visited):
        if srcTblElt is not None:
            for rel in tblAxisRelSet.fromModelObject(srcTblElt):
                srcAxisElt = rel.toModelObject
                if isinstance(srcAxisElt, ModelEuAxisCoord):
                    visited.add(srcAxisElt)
                    newAxisElt = etree.SubElement(newLinkElt, "{http://xbrl.org/2011/table}ruleAxis")
                    copyAttrs(srcAxisElt, newAxisElt, ("id",
                                                       "{http://www.w3.org/1999/xlink}type",
                                                       "{http://www.w3.org/1999/xlink}label"))
                    newAxisElt.set("abstract", "true") # always true on root element
                    newArcElt = etree.SubElement(newLinkElt, "{http://xbrl.org/2011/table}axisArc")
                    copyAttrs(rel, newArcElt, ("id",
                                               "{http://www.w3.org/1999/xlink}type",
                                               "{http://www.w3.org/1999/xlink}from",
                                               "{http://www.w3.org/1999/xlink}to",
                                               "order"))
                    newArcElt.set("{http://www.w3.org/1999/xlink}arcrole", XbrlConst.tableBreakdown)
                    newArcElt.set("axisDisposition", rel.axisDisposition)
                    generateAxis(newLinkElt, newAxisElt, srcAxisElt, axisMbrRelSet, visited)
                    visited.discard(srcAxisElt)

    def generateAxis(newLinkElt, newAxisParentElt, srcAxisElt, axisMbrRelSet, visited):
        for rel in axisMbrRelSet.fromModelObject(srcAxisElt):
            tgtAxisElt = rel.toModelObject
            if isinstance(tgtAxisElt, ModelEuAxisCoord) and tgtAxisElt not in visited:
                visited.add(tgtAxisElt)
                newAxisElt = etree.SubElement(newLinkElt, "{http://xbrl.org/2011/table}ruleAxis")
                copyAttrs(tgtAxisElt, newAxisElt, ("id",
                                                   "abstract",
                                                   "{http://www.w3.org/1999/xlink}type",
                                                   "{http://www.w3.org/1999/xlink}label"))
                if tgtAxisElt.primaryItemQname:
                    newRuleElt = etree.SubElement(newAxisElt, "{http://xbrl.org/2008/formula}concept")
                    newQnameElt = etree.SubElement(newRuleElt, "{http://xbrl.org/2008/formula}qname")
                    newQnameElt.text = XmlUtil.addQnameValue(docObj, tgtAxisElt.primaryItemQname)
                for dimQname, memQname in tgtAxisElt.explicitDims:
                    newRuleElt = etree.SubElement(newAxisElt, "{http://xbrl.org/2008/formula}explicitDimension")
                    newRuleElt.set("dimension", XmlUtil.addQnameValue(docObj, dimQname))
                    newMbrElt = etree.SubElement(newRuleElt, "{http://xbrl.org/2008/formula}member")
                    newQnameElt = etree.SubElement(newMbrElt, "{http://xbrl.org/2008/formula}qname")
                    newQnameElt.text = XmlUtil.addQnameValue(docObj, memQname)
                newArcElt = etree.SubElement(newLinkElt, "{http://xbrl.org/2011/table}axisArc")
                copyAttrs(rel, newArcElt, ("id",
                                           "{http://www.w3.org/1999/xlink}type",
                                           "{http://www.w3.org/1999/xlink}from",
                                           "{http://www.w3.org/1999/xlink}to",
                                           "order"))
                newArcElt.set("{http://www.w3.org/1999/xlink}arcrole", XbrlConst.tableAxisSubtree)
                generateAxis(newLinkElt, newAxisElt, tgtAxisElt, axisMbrRelSet, visited)
                visited.discard(tgtAxisElt)

    # sort URIs
    linkroleUris = sorted([linkroleUri
                           for linkroleUri in tblAxisRelSet.linkRoleUris])

    firstNewLinkElt = None
    roleRefUris = set()
    for linkroleUri in linkroleUris:
        numELRs += 1
        newLinkElt = etree.SubElement(lbElement, "{http://xbrl.org/2008/generic}link")
        newLinkElt.set("{http://www.w3.org/1999/xlink}type", "extended")
        newLinkElt.set("{http://www.w3.org/1999/xlink}role", linkroleUri)
        if firstNewLinkElt is None: firstNewLinkElt = newLinkElt
        # To do: add roleRef if needed
        tblAxisRelSet = dts.relationshipSet(XbrlConst.euTableAxis, linkroleUri)
        axisMbrRelSet = dts.relationshipSet(XbrlConst.euAxisMember, linkroleUri)
        for srcTblElt in tblAxisRelSet.rootConcepts:
            if srcTblElt.tag == "{http://www.eurofiling.info/2010/rendering}table":
                numTables += 1
                newTblElt = etree.SubElement(newLinkElt, "{http://xbrl.org/2011/table}table")
                newTblElt.set("aspectModel", "dimensional")
                copyAttrs(srcTblElt, newTblElt, ("id",
                                                 "{http://www.w3.org/1999/xlink}type",
                                                 "{http://www.w3.org/1999/xlink}label"))
                generateTable(newLinkElt, newTblElt, srcTblElt, tblAxisRelSet, axisMbrRelSet, set())

                if linkroleUri not in roleRefUris:
                    srcRoleRefElt = XmlUtil.descendant(srcTblElt.getroottree(), XbrlConst.link, "roleRef", "roleURI", linkroleUri)
                    if srcRoleRefElt is not None:
                        roleRefUris.add(linkroleUri)
                        newRoleRefElt = etree.Element("{http://www.xbrl.org/2003/linkbase}roleRef")
                        copyAttrs(srcRoleRefElt, newRoleRefElt, ("roleURI",
                                                                 "{http://www.w3.org/1999/xlink}type",
                                                                 "{http://www.w3.org/1999/xlink}href"))
                        firstNewLinkElt.addprevious(newRoleRefElt)

    fh = open(updatedTableLinkbaseFile, "w", encoding="utf-8")
    XmlUtil.writexml(fh, xmlDocument, encoding="utf-8")
    fh.close()

    dts.info("info:updateTableLinkbase",
             _("Updated Table Linkbase of %(entryFile)s has %(numberOfLinkroles)s linkroles, %(numberOfTables)s tables in file %(tableLBOutputFile)s."),
             modelObject=dts,
             entryFile=dts.uri, numberOfLinkroles=numELRs, numberOfTables=numTables, tableLBOutputFile=updatedTableLinkbaseFile)

def updateTableLBMenuEntender(cntlr, menu):
    # Extend menu with an item for the savedts plugin
    menu.add_command(label="Update table linkbase",
                     underline=0,
                     command=lambda: updateTableLBMenuCommand(cntlr) )

def updateTableLBMenuCommand(cntlr):
    # save DTS menu item has been invoked
    if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
        cntlr.addToLog("No taxonomy loaded.")
        return

        # get file name into which to save log file while in foreground thread
    updatedTableLinkbaseFile = cntlr.uiFileDialog("save",
                                                  title=_("arelle - Save Updated Table Linkbase file"),
            initialdir=cntlr.config.setdefault("tableLinkbaseFileDir","."),
            filetypes=[(_("Linkbase file .xml"), "*.xml")],
            defaultextension=".xml")
    if not updatedTableLinkbaseFile:
        return False
    import os
    cntlr.config["tableLinkbaseFileDir"] = os.path.dirname(updatedTableLinkbaseFile)
    cntlr.saveConfig()

    try:
        generateUpdatedTableLB(cntlr.modelManager.modelXbrl, updatedTableLinkbaseFile)
    except Exception as ex:
        dts = cntlr.modelManager.modelXbrl
        dts.error("exception",
            _("Updated table linkbase generation exception: %(error)s"), error=ex,
            modelXbrl=dts,
            exc_info=True)

def updateTableLBCommandLineOptionExtender(parser):
    # extend command line options with a save DTS option
    parser.add_option("--save-updated-table-linkbase",
                      action="store",
                      dest="updatedTableLinkbaseFile",
                      help=_("Update table linkbase file from Eurofiling 2010 to XII 2011 syntax."))

def updateTableLBCommandLineXbrlRun(cntlr, options, modelXbrl, *args, **kwargs):
    # extend XBRL-loaded run processing for this option
    if getattr(options, "updatedTableLinkbaseFile", None):
        if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
            cntlr.addToLog("No taxonomy loaded.")
            return
        generateUpdatedTableLB(cntlr.modelManager.modelXbrl, options.updatedTableLinkbaseFile)


__pluginInfo__ = {
    'name': 'Update Table Linkbase',
    'version': '0.9',
    'description': "This plug-in updates a table linkbase from Eurofiling 2010 syntax to XII 2011 PWD syntax. "
                   "The updated linkbase is saved to a local file.  "
                   "Hrefs in the new file are identical the prior hrefs (not offset considering save-to directory).  ",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'CntlrWinMain.Menu.Tools': updateTableLBMenuEntender,
    'CntlrCmdLine.Options': updateTableLBCommandLineOptionExtender,
    'CntlrCmdLine.Xbrl.Run': updateTableLBCommandLineXbrlRun,
}



'''
See COPYRIGHT.md for copyright information.
'''
import os
from arelle import ModelDocument, ModelValue, XmlUtil
from arelle.ModelValue import qname
from arelle.plugin.validate.EFM.Document import checkDTSdocument
from arelle.plugin.validate.EFM.Filing import validateFiling
from arelle.Version import authorLabel, copyrightLabel

def dislosureSystemTypes(disclosureSystem, *args, **kwargs):
    # return ((disclosure system name, variable name), ...)
    return (("GFM", "GFMplugin"),)

def disclosureSystemConfigURL(disclosureSystem, *args, **kwargs):
    return os.path.join(os.path.dirname(__file__), "config.xml")

def validateXbrlStart(val, parameters=None, *args, **kwargs):
    val.validateGFMplugin = val.validateDisclosureSystem and getattr(val.disclosureSystem, "GFMplugin", False)
    if not (val.validateGFMplugin):
        return

def validateXbrlFinally(val, *args, **kwargs):
    if not (val.validateGFMplugin):
        return

    modelXbrl = val.modelXbrl

    _statusMsg = _("validating {0} filing rules").format(val.disclosureSystem.name)
    modelXbrl.profileActivity()
    modelXbrl.modelManager.showStatus(_statusMsg)

    validateFiling(val, modelXbrl, isGFM=True)

    modelXbrl.profileActivity(_statusMsg, minTimeToShow=0.0)
    modelXbrl.modelManager.showStatus(None)

def validateXbrlDtsDocument(val, modelDocument, isFilingDocument, *args, **kwargs):
    if not (val.validateGFMplugin) or not modelDocument:
        return

    checkDTSdocument(val, modelDocument, isFilingDocument)


__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Validate GFM',
    'version': '0.9',
    'description': '''GFM Validation.''',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'DisclosureSystem.Types': dislosureSystemTypes,
    'DisclosureSystem.ConfigURL': disclosureSystemConfigURL,
    'Validate.XBRL.Start': validateXbrlStart,
    'Validate.XBRL.Finally': validateXbrlFinally,
    'Validate.XBRL.DTS.document': validateXbrlDtsDocument,
}



'''

Use this module to start Arelle in command line modes

See COPYRIGHT.md for copyright information.
'''
import os
import sys

# Only imports for modules distributed as part of the standard Python library may go above this line.
# All other imports will result in module not found exceptions on the frozen macOS build.
if sys.platform == "darwin" and getattr(sys, 'frozen', False):
    for i in range(len(sys.path)):  # signed code can't contain python modules
        sys.path.append(sys.path[i].replace("MacOS", "Resources"))

import regex as re

from arelle.SocketUtils import INTERNET_CONNECTIVITY, OFFLINE, warnSocket

internetConnectivityArgPattern = rf'--({INTERNET_CONNECTIVITY}|{INTERNET_CONNECTIVITY.lower()})'
internetConnectivityArgRegex = re.compile(internetConnectivityArgPattern)
internetConnectivityOfflineEqualsRegex = re.compile(f"{internetConnectivityArgPattern}={OFFLINE}")
prevArg = ''
for arg in sys.argv[1:]:
    if (
            internetConnectivityOfflineEqualsRegex.fullmatch(arg)
            or (internetConnectivityArgRegex.fullmatch(prevArg) and arg == OFFLINE)
    ):
        warnSocket()
        break
    prevArg = arg

from arelle.BetaFeatures import BETA_OBJECT_MODEL_FEATURE, enableNewObjectModel

# Model transition must be enabled before any other imports to avoid mixing base classes.
if f"--{BETA_OBJECT_MODEL_FEATURE}" in sys.argv or f"--{BETA_OBJECT_MODEL_FEATURE.lower()}" in sys.argv:
    enableNewObjectModel()

from arelle import CntlrCmdLine, CntlrComServer

if '--COMserver' in sys.argv:
    CntlrComServer.main()
elif __name__.startswith('_mod_wsgi_') or os.getenv('wsgi.version'):
    application = CntlrCmdLine.wsgiApplication()
elif __name__ in ('__main__', 'arelleCmdLine__main__', 'arellecmdline__main__'): #cx_Freeze 5 prepends module name to __main__
    CntlrCmdLine.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals

import sys

from tatsu.buffering import Buffer
from tatsu.parsing import Parser
from tatsu.parsing import tatsumasu
from tatsu.util import re, generic_main


KEYWORDS = {}  # type: ignore


class XFBuffer(Buffer):
    def __init__(
        self,
        text,
        whitespace=None,
        nameguard=None,
        comments_re='(?s)\\(:((?!(\\(:|:\\))).)*:\\)',
        eol_comments_re=None,
        ignorecase=None,
        namechars='-',
        **kwargs
    ):
        super(XFBuffer, self).__init__(
            text,
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            namechars=namechars,
            **kwargs
        )


class XFParser(Parser):
    def __init__(
        self,
        whitespace=None,
        nameguard=None,
        comments_re='(?s)\\(:((?!(\\(:|:\\))).)*:\\)',
        eol_comments_re=None,
        ignorecase=None,
        left_recursion=True,
        parseinfo=True,
        keywords=None,
        namechars='-',
        buffer_class=XFBuffer,
        **kwargs
    ):
        if keywords is None:
            keywords = KEYWORDS
        super(XFParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            left_recursion=left_recursion,
            parseinfo=parseinfo,
            keywords=keywords,
            namechars=namechars,
            buffer_class=buffer_class,
            **kwargs
        )

    @tatsumasu()
    def _module_(self):

        def block0():
            self._namespace_declaration_()
        self._closure(block0)

        def block1():
            self._default_()
        self._closure(block1)

        def block2():
            self._parameter_()
        self._closure(block2)

        def block3():
            with self._choice():
                with self._option():
                    self._filter_declaration_()
                with self._option():
                    self._fact_variable_()
                with self._option():
                    self._general_variable_()
                with self._option():
                    self._function_declaration_()
                self._error('no available options')
        self._closure(block3)

        def block5():
            self._assertion_()
        self._closure(block5)
        self._check_eof()

    @tatsumasu()
    def _separator_(self):
        self._token(';')

    @tatsumasu()
    def _namespace_declaration_(self):
        self._token('namespace')
        self._name_()
        self._token('=')
        self._quoted_url_()
        self._separator_()

    @tatsumasu()
    def _default_(self):
        with self._choice():
            with self._option():
                self._severity_()
            with self._option():
                self._token('default-language')
                self._language_()
                self._separator_()
            self._error('no available options')

    @tatsumasu()
    def _severity_(self):
        self._token('unsatisfied-severity')
        self._message_severity_()
        self._separator_()

    @tatsumasu()
    def _message_severity_(self):
        with self._choice():
            with self._option():
                self._token('ERROR')
            with self._option():
                self._token('WARNING')
            with self._option():
                self._token('INFO')
            self._error('no available options')

    @tatsumasu()
    def _parameter_(self):
        self._token('parameter')
        self._name_()
        with self._optional():
            self._token('required')
        self._token('{')
        with self._optional():
            self._token('select')
            self._enclosed_expression_()
        with self._optional():
            self._token('as')
            self._qname_()
        self._token('}')
        self._separator_()

    @tatsumasu()
    def _filter_declaration_(self):
        self._token('filter')
        self._name_()
        self._token('{')

        def block0():
            self._filter_()
        self._positive_closure(block0)
        self._token('}')
        self._separator_()

    @tatsumasu()
    def _filter_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._concept_filter_()
                with self._option():
                    self._general_filter_()
                with self._option():
                    self._period_filter_()
                with self._option():
                    self._dimension_filter_()
                with self._option():
                    self._unit_filter_()
                with self._option():
                    self._entity_filter_()
                with self._option():
                    self._match_filter_()
                with self._option():
                    self._relative_filter_()
                with self._option():
                    self._tuple_filter_()
                with self._option():
                    self._value_filter_()
                with self._option():
                    self._boolean_filter_()
                with self._option():
                    self._aspect_cover_filter_()
                with self._option():
                    self._concept_relation_filter_()
                with self._option():
                    self._declared_filter_reference_()
                self._error('no available options')

    @tatsumasu()
    def _concept_filter_(self):
        with self._optional():
            self._token('complemented')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('covering')
                with self._option():
                    self._constant('covering')
                with self._option():
                    self._token('non-covering')
                self._error('no available options')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('concept-name')
                    self._cut()

                    def block1():
                        with self._choice():
                            with self._option():
                                self._qname_()
                            with self._option():
                                self._localname_()
                            with self._option():
                                self._enclosed_expression_()
                            self._error('no available options')
                    self._positive_closure(block1)
                with self._option():
                    self._token('concept-period-type')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._token('instant')
                            with self._option():
                                self._token('duration')
                            self._error('no available options')
                with self._option():
                    self._token('concept-balance')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._token('credit')
                            with self._option():
                                self._token('debit')
                            with self._option():
                                self._token('none')
                            self._error('no available options')
                with self._option():
                    self._token('concept-data-type')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._token('strict')
                            with self._option():
                                self._token('non-strict')
                            self._error('no available options')
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._qname_()
                            with self._option():
                                self._enclosed_expression_()
                            self._error('no available options')
                with self._option():
                    self._token('concept-substitution-group')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._token('strict')
                            with self._option():
                                self._token('non-strict')
                            self._error('no available options')
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._qname_()
                            with self._option():
                                self._enclosed_expression_()
                            self._error('no available options')
                self._error('no available options')
        self._separator_()

    @tatsumasu()
    def _general_filter_(self):
        with self._optional():
            self._token('complemented')
        self._token('general')
        self._cut()
        self._enclosed_expression_()
        self._separator_()

    @tatsumasu()
    def _period_filter_(self):
        with self._optional():
            self._token('complemented')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('period-start')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._date_constant_()
                            with self._option():
                                self._date_time_constant_()
                            with self._option():
                                with self._group():
                                    self._token('date')
                                    self._enclosed_expression_()
                                    with self._optional():
                                        self._token('time')
                                        self._enclosed_expression_()
                            self._error('no available options')
                with self._option():
                    self._token('period-end')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._date_constant_()
                            with self._option():
                                self._date_time_constant_()
                            with self._option():
                                with self._group():
                                    self._token('date')
                                    self._enclosed_expression_()
                                    with self._optional():
                                        self._token('time')
                                        self._enclosed_expression_()
                            self._error('no available options')
                with self._option():
                    self._token('period-instant')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._date_constant_()
                            with self._option():
                                self._date_time_constant_()
                            with self._option():
                                with self._group():
                                    self._token('date')
                                    self._enclosed_expression_()
                                    with self._optional():
                                        self._token('time')
                                        self._enclosed_expression_()
                            self._error('no available options')
                with self._option():
                    self._token('period')
                    self._cut()
                    self._enclosed_expression_()
                with self._option():
                    self._token('instant-duration')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._token('start')
                            with self._option():
                                self._token('end')
                            self._error('no available options')
                    self._variable_ref_()
                self._error('no available options')
        self._separator_()

    @tatsumasu()
    def _dimension_filter_(self):
        with self._optional():
            self._token('complemented')
        with self._group():
            with self._choice():
                with self._option():
                    with self._group():
                        self._token('explicit-dimension')
                        self._cut()
                        with self._group():
                            with self._choice():
                                with self._option():
                                    self._variable_ref_()
                                with self._option():
                                    self._qname_()
                                with self._option():
                                    self._localname_()
                                with self._option():
                                    self._enclosed_expression_()
                                self._error('no available options')

                        def block1():
                            with self._choice():
                                with self._option():
                                    self._token('default-member')
                                with self._option():
                                    self._token('member')
                                    self._cut()
                                    with self._group():
                                        with self._choice():
                                            with self._option():
                                                self._variable_ref_()
                                            with self._option():
                                                self._qname_()
                                            with self._option():
                                                self._localname_()
                                            with self._option():
                                                self._enclosed_expression_()
                                            self._error('no available options')
                                    with self._optional():
                                        self._token('linkrole')
                                        self._quoted_anyURI_()
                                        self._token('arcrole')
                                        self._quoted_anyURI_()
                                        self._token('axis')
                                        self._dimension_axis_()
                                self._error('no available options')
                        self._closure(block1)
                with self._option():
                    with self._group():
                        self._token('typed-dimension')
                        self._cut()
                        with self._group():
                            with self._choice():
                                with self._option():
                                    self._variable_ref_()
                                with self._option():
                                    self._qname_()
                                with self._option():
                                    self._localname_()
                                with self._option():
                                    self._enclosed_expression_()
                                self._error('no available options')
                        with self._optional():
                            self._token('test')
                            self._enclosed_expression_()
                self._error('no available options')
        self._separator_()

    @tatsumasu()
    def _dimension_axis_(self):
        with self._choice():
            with self._option():
                self._token('child')
            with self._option():
                self._token('child-or-self')
            with self._option():
                self._token('descendant')
            with self._option():
                self._token('descendant-or-self')
            self._error('no available options')

    @tatsumasu()
    def _unit_filter_(self):
        with self._optional():
            self._token('complemented')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('unit-single-measure')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._qname_()
                            with self._option():
                                self._enclosed_expression_()
                            self._error('no available options')
                with self._option():
                    self._token('unit-general-measures')
                    self._cut()
                    self._enclosed_expression_()
                self._error('no available options')
        self._separator_()

    @tatsumasu()
    def _entity_filter_(self):
        with self._optional():
            self._token('complemented')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('entity-scheme-pattern')
                    self._cut()
                    self._regexp_pattern_()
                with self._option():
                    self._token('entity-scheme')
                    self._cut()
                    self._enclosed_expression_()
                with self._option():
                    self._token('entity-identifier-pattern')
                    self._cut()
                    self._regexp_pattern_()
                with self._option():
                    self._token('entity-identifier')
                    self._cut()
                    self._enclosed_expression_()
                with self._option():
                    self._token('entity')
                    self._cut()
                    self._token('scheme')
                    self._enclosed_expression_()
                    self._token('value')
                    self._enclosed_expression_()
                self._error('no available options')
        self._separator_()

    @tatsumasu()
    def _match_filter_(self):
        with self._optional():
            self._token('complemented')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('match-concept')
                    self._cut()
                    self._variable_ref_()
                with self._option():
                    self._token('match-location')
                    self._cut()
                    self._variable_ref_()
                with self._option():
                    self._token('match-entity-identifier')
                    self._cut()
                    self._variable_ref_()
                with self._option():
                    self._token('match-period')
                    self._cut()
                    self._variable_ref_()
                with self._option():
                    self._token('match-unit')
                    self._cut()
                    self._variable_ref_()
                with self._option():
                    self._token('match-dimension')
                    self._cut()
                    self._variable_ref_()
                    self._token('dimension')
                    self._qname_()
                self._error('no available options')
        with self._optional():
            self._token('match-any')
        self._separator_()

    @tatsumasu()
    def _relative_filter_(self):
        with self._optional():
            self._token('complemented')
        self._token('relative')
        self._cut()
        self._variable_ref_()
        self._separator_()

    @tatsumasu()
    def _tuple_filter_(self):
        with self._optional():
            self._token('complemented')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('parent')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._qname_()
                            with self._option():
                                self._enclosed_expression_()
                            self._error('no available options')
                with self._option():
                    self._token('ancestor')
                    self._cut()
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._qname_()
                            with self._option():
                                self._enclosed_expression_()
                            self._error('no available options')
                with self._option():
                    self._token('sibling')
                    self._cut()
                    self._variable_ref_()
                self._error('no available options')
        self._separator_()

    @tatsumasu()
    def _value_filter_(self):
        with self._optional():
            self._token('complemented')
        self._token('nilled')
        self._separator_()

    @tatsumasu()
    def _boolean_filter_(self):
        with self._optional():
            self._token('complemented')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('and')
                with self._option():
                    self._token('or')
                self._error('no available options')
        self._cut()
        self._token('{')

        def block1():
            self._filter_()
        self._positive_closure(block1)
        self._token('}')
        self._separator_()

    @tatsumasu()
    def _aspect_cover_filter_(self):
        with self._optional():
            self._token('complemented')
        self._token('aspect-cover')
        self._cut()

        def block0():
            with self._choice():
                with self._option():
                    self._token('all')
                with self._option():
                    self._token('concept')
                with self._option():
                    self._token('entity-identifier')
                with self._option():
                    self._token('location')
                with self._option():
                    self._token('period')
                with self._option():
                    self._token('unit')
                with self._option():
                    self._token('dimensions')
                with self._option():
                    self._token('dimension')
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._qname_()
                            with self._option():
                                self._localname_()
                            with self._option():
                                self._enclosed_expression_()
                            self._error('no available options')
                with self._option():
                    self._token('exclude-dimension')
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._qname_()
                            with self._option():
                                self._localname_()
                            with self._option():
                                self._enclosed_expression_()
                            self._error('no available options')
                self._error('no available options')
        self._positive_closure(block0)
        self._separator_()

    @tatsumasu()
    def _concept_relation_filter_(self):
        with self._optional():
            self._token('complemented')
        self._token('concept-relation')
        self._cut()
        with self._group():
            with self._choice():
                with self._option():
                    self._variable_ref_()
                with self._option():
                    self._qname_()
                with self._option():
                    self._localname_()
                with self._option():
                    self._enclosed_expression_()
                self._error('no available options')
        with self._optional():
            self._token('linkrole')
            with self._group():
                with self._choice():
                    with self._option():
                        self._quoted_anyURI_()
                    with self._option():
                        self._enclosed_expression_()
                    self._error('no available options')
        with self._optional():
            self._token('arcrole')
            with self._group():
                with self._choice():
                    with self._option():
                        self._quoted_anyURI_()
                    with self._option():
                        self._enclosed_expression_()
                    self._error('no available options')
        with self._optional():
            self._token('axis')
            self._relation_axis_()
        with self._optional():
            self._token('generations')
            self._non_negative_integer_()
        with self._optional():
            self._token('test')
            self._enclosed_expression_()
        self._separator_()

    @tatsumasu()
    def _relation_axis_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._token('child-or-self')
                with self._option():
                    self._token('child')
                with self._option():
                    self._token('descendant-or-self')
                with self._option():
                    self._token('descendant')
                with self._option():
                    self._token('parent-or-self')
                with self._option():
                    self._token('parent')
                with self._option():
                    self._token('ancestor-or-self')
                with self._option():
                    self._token('ancestor')
                with self._option():
                    self._token('sibling-or-self')
                with self._option():
                    self._token('sibling-or-descendant')
                with self._option():
                    self._token('sibling')
                self._error('no available options')
        self._separator_()

    @tatsumasu()
    def _declared_filter_reference_(self):
        self._token('filter')
        self._cut()
        self._variable_ref_()
        self._separator_()

    @tatsumasu()
    def _fact_variable_(self):
        self._token('variable')
        self._cut()
        self._variable_ref_()
        self._token('{')
        with self._optional():
            self._token('bind-as-sequence')
        with self._optional():
            self._token('nils')
        with self._optional():
            self._token('matches')
        with self._optional():
            self._token('fallback')
            self._enclosed_expression_()

        def block0():
            self._filter_()
        self._closure(block0)
        self._token('}')
        self._separator_()

    @tatsumasu()
    def _general_variable_(self):
        self._token('variable')
        self._cut()
        self._variable_ref_()
        self._token('{')
        with self._optional():
            self._token('bind-as-sequence')
        self._token('select')
        self._enclosed_expression_()
        self._token('}')
        self._separator_()

    @tatsumasu()
    def _function_declaration_(self):
        self._token('function')
        self._cut()
        self._qname_()
        self._token('(')

        def block0():
            self._NCNAME_FRAG_()
            self._token('as')
            self._qname_()
        self._closure(block0)
        self._token(')')
        self._token('as')
        self._qname_()
        with self._optional():
            self._token('{')
            self._token('return')
            self._enclosed_expression_()
            self._separator_()
            self._token('}')
        self._separator_()

    @tatsumasu()
    def _assertion_(self):
        self._token('assertion')
        self._name_()
        self._token('{')

        def block0():
            with self._choice():
                with self._option():
                    self._label_()
                with self._option():
                    self._message_()
                with self._option():
                    self._severity_()
                with self._option():
                    self._token('aspect-model-non-dimensional')
                    self._separator_()
                with self._option():
                    self._token('no-implicit-filtering')
                    self._separator_()
                self._error('no available options')
        self._closure(block0)

        def block2():
            self._filter_()
        self._closure(block2)

        def block3():
            with self._choice():
                with self._option():
                    self._fact_variable_()
                with self._option():
                    self._general_variable_()
                with self._option():
                    self._referenced_parameter_()
                self._error('no available options')
        self._closure(block3)

        def block5():
            self._precondition_()
        self._closure(block5)
        with self._group():
            with self._choice():
                with self._option():
                    self._value_expression_()
                with self._option():
                    self._existence_expression_()
                self._error('no available options')
        self._token('}')
        self._separator_()

    @tatsumasu()
    def _label_(self):
        self._token('label')
        with self._optional():
            self._token('(')
            self._language_()
            self._token(')')
        self._quoted_string_()
        self._separator_()

    @tatsumasu()
    def _message_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._token('satisfied-message')
                with self._option():
                    self._token('unsatisfied-message')
                self._error('no available options')
        self._cut()
        with self._optional():
            self._token('(')
            self._cut()
            self._language_()
            self._token(')')
        self._quoted_string_()
        self._separator_()

    @tatsumasu()
    def _referenced_parameter_(self):
        self._token('parameter')
        self._cut()
        self._variable_ref_()
        self._token('references')
        self._qname_()
        self._separator_()

    @tatsumasu()
    def _precondition_(self):
        self._token('precondition')
        self._enclosed_expression_()

    @tatsumasu()
    def _value_expression_(self):
        self._token('test')
        self._enclosed_expression_()
        self._separator_()

    @tatsumasu()
    def _existence_expression_(self):
        self._token('evaluation-count')
        self._enclosed_expression_()
        self._separator_()

    @tatsumasu()
    def _enclosed_expression_(self):
        self._token('{')
        self._xPath_()
        self._token('}')

    @tatsumasu()
    def _name_(self):
        self._NCNAME_FRAG_()

    @tatsumasu()
    def _language_(self):
        self._pattern(r'[A-Za-z]{2}(-[A-Za-z]{2})?')

    @tatsumasu()
    def _quoted_anyURI_(self):
        self._quoted_url_()

    @tatsumasu()
    def _quoted_url_(self):
        self._quoted_string_()

    @tatsumasu()
    def _localname_(self):
        self._name_()

    @tatsumasu()
    def _variable_ref_(self):
        self._token('$')
        self._variable_name_()

    @tatsumasu()
    def _variable_name_(self):
        self._name_()

    @tatsumasu()
    def _non_negative_integer_(self):
        self._pattern(r'[0-9]+')

    @tatsumasu()
    def _date_time_constant_(self):
        self._pattern(r'\b(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\\.[0-9]+)?(Z)?\b')

    @tatsumasu()
    def _date_constant_(self):
        self._pattern(r'\b(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])\b')

    @tatsumasu()
    def _quoted_string_(self):
        self._pattern(r'("([^\\"]|\\.)*"|\'([^\\\']|\\.)*\')')

    @tatsumasu()
    def _regexp_pattern_(self):
        self._pattern(r'\/([^\\\/]|\\.)*\/')

    @tatsumasu()
    def _xPath_(self):
        self._expr_()

    @tatsumasu()
    def _expr_(self):
        self._exprSingle_()

        def block0():
            self._token(',')
            self._exprSingle_()
        self._closure(block0)

    @tatsumasu()
    def _exprSingle_(self):
        with self._choice():
            with self._option():
                self._forExpr_()
            with self._option():
                self._quantifiedExpr_()
            with self._option():
                self._ifExpr_()
            with self._option():
                self._orExpr_()
            self._error('no available options')

    @tatsumasu()
    def _forExpr_(self):
        self._simpleForClause_()
        self._token('return')
        self._exprSingle_()

    @tatsumasu()
    def _simpleForClause_(self):
        self._token('for')
        self._token('$')
        self._varName_()
        self._token('in')
        self._exprSingle_()

        def block0():
            self._token(',')
            self._token('$')
            self._varName_()
            self._token('in')
            self._exprSingle_()
        self._closure(block0)

    @tatsumasu()
    def _quantifiedExpr_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._token('some')
                with self._option():
                    self._token('every')
                self._error('no available options')
        self._token('$')
        self._varName_()
        self._token('in')
        self._exprSingle_()

        def block1():
            self._token(',')
            self._token('$')
            self._varName_()
            self._token('in')
            self._exprSingle_()
        self._closure(block1)
        self._token('satisfies')
        self._exprSingle_()

    @tatsumasu()
    def _ifExpr_(self):
        self._token('if')
        self._token('(')
        self._expr_()
        self._token(')')
        self._token('then')
        self._exprSingle_()
        self._token('else')
        self._exprSingle_()

    @tatsumasu()
    def _orExpr_(self):
        self._andExpr_()

        def block0():
            self._token('or')
            self._andExpr_()
        self._closure(block0)

    @tatsumasu()
    def _andExpr_(self):
        self._comparisonExpr_()

        def block0():
            self._token('and')
            self._comparisonExpr_()
        self._closure(block0)

    @tatsumasu()
    def _comparisonExpr_(self):
        self._rangeExpr_()
        with self._optional():
            with self._group():
                with self._choice():
                    with self._option():
                        self._valueComp_()
                    with self._option():
                        self._nodeComp_()
                    with self._option():
                        self._generalComp_()
                    self._error('no available options')
            self._rangeExpr_()

    @tatsumasu()
    def _rangeExpr_(self):
        self._additiveExpr_()
        with self._optional():
            self._token('to')
            self._additiveExpr_()

    @tatsumasu()
    def _additiveExpr_(self):
        self._multiplicativeExpr_()

        def block0():
            with self._group():
                with self._choice():
                    with self._option():
                        self._token('+')
                    with self._option():
                        self._token('-')
                    self._error('no available options')
            self._multiplicativeExpr_()
        self._closure(block0)

    @tatsumasu()
    def _multiplicativeExpr_(self):
        self._unionExpr_()

        def block0():
            with self._group():
                with self._choice():
                    with self._option():
                        self._token('*')
                    with self._option():
                        self._token('div')
                    with self._option():
                        self._token('idiv')
                    with self._option():
                        self._token('mod')
                    self._error('no available options')
            self._unionExpr_()
        self._closure(block0)

    @tatsumasu()
    def _unionExpr_(self):
        self._intersectExceptExpr_()

        def block0():
            with self._group():
                with self._choice():
                    with self._option():
                        self._token('union')
                    with self._option():
                        self._token('|')
                    self._error('no available options')
            self._intersectExceptExpr_()
        self._closure(block0)

    @tatsumasu()
    def _intersectExceptExpr_(self):
        self._instanceofExpr_()

        def block0():
            with self._group():
                with self._choice():
                    with self._option():
                        self._token('intersect')
                    with self._option():
                        self._token('except')
                    self._error('no available options')
            self._instanceofExpr_()
        self._closure(block0)

    @tatsumasu()
    def _instanceofExpr_(self):
        self._treatExpr_()
        with self._optional():
            self._token('instance')
            self._token('of')
            self._sequenceType_()

    @tatsumasu()
    def _treatExpr_(self):
        self._castableExpr_()
        with self._optional():
            self._token('treat')
            self._token('as')
            self._sequenceType_()

    @tatsumasu()
    def _castableExpr_(self):
        self._castExpr_()
        with self._optional():
            self._token('castable')
            self._token('as')
            self._singleType_()

    @tatsumasu()
    def _castExpr_(self):
        self._unaryExpr_()
        with self._optional():
            self._token('cast')
            self._token('as')
            self._singleType_()

    @tatsumasu()
    def _unaryExpr_(self):

        def block0():
            with self._choice():
                with self._option():
                    self._token('-')
                with self._option():
                    self._token('+')
                self._error('no available options')
        self._closure(block0)
        self._valueExpr_()

    @tatsumasu()
    def _valueExpr_(self):
        self._pathExpr_()

    @tatsumasu()
    def _generalComp_(self):
        with self._choice():
            with self._option():
                self._token('=')
            with self._option():
                self._token('!=')
            with self._option():
                self._token('<=')
            with self._option():
                self._token('<')
            with self._option():
                self._token('>=')
            with self._option():
                self._token('>')
            self._error('no available options')

    @tatsumasu()
    def _valueComp_(self):
        with self._choice():
            with self._option():
                self._token('eq')
            with self._option():
                self._token('ne')
            with self._option():
                self._token('lt')
            with self._option():
                self._token('le')
            with self._option():
                self._token('gt')
            with self._option():
                self._token('ge')
            self._error('no available options')

    @tatsumasu()
    def _nodeComp_(self):
        with self._choice():
            with self._option():
                self._token('is')
            with self._option():
                self._token('<<')
            with self._option():
                self._token('>>')
            self._error('no available options')

    @tatsumasu()
    def _pathExpr_(self):
        with self._choice():
            with self._option():
                with self._group():
                    self._token('//')
                    self._relativePathExpr_()
            with self._option():
                with self._group():
                    self._token('/')
                    with self._optional():
                        self._relativePathExpr_()
            with self._option():
                self._relativePathExpr_()
            self._error('no available options')

    @tatsumasu()
    def _relativePathExpr_(self):
        self._stepExpr_()

        def block0():
            with self._group():
                with self._choice():
                    with self._option():
                        self._token('/')
                    with self._option():
                        self._token('//')
                    self._error('no available options')
            self._stepExpr_()
        self._closure(block0)

    @tatsumasu()
    def _stepExpr_(self):
        with self._choice():
            with self._option():
                self._filterExpr_()
            with self._option():
                self._axisStep_()
            self._error('no available options')

    @tatsumasu()
    def _axisStep_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._reverseStep_()
                with self._option():
                    self._forwardStep_()
                self._error('no available options')
        self._predicateList_()

    @tatsumasu()
    def _forwardStep_(self):
        with self._choice():
            with self._option():
                with self._group():
                    self._forwardAxis_()
                    self._nodeTest_()
            with self._option():
                self._abbrevForwardStep_()
            self._error('no available options')

    @tatsumasu()
    def _forwardAxis_(self):
        with self._choice():
            with self._option():
                with self._group():
                    self._token('child')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('descendant')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('attribute')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('self')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('descendant-or-self')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('following-sibling')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('following')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('namespace')
                    self._token('::')
            self._error('no available options')

    @tatsumasu()
    def _abbrevForwardStep_(self):
        with self._optional():
            self._token('@')
        self._nodeTest_()

    @tatsumasu()
    def _reverseStep_(self):
        with self._choice():
            with self._option():
                with self._group():
                    self._reverseAxis_()
                    self._nodeTest_()
            with self._option():
                self._abbrevReverseStep_()
            self._error('no available options')

    @tatsumasu()
    def _reverseAxis_(self):
        with self._choice():
            with self._option():
                with self._group():
                    self._token('parent')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('ancestor')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('preceding-sibling')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('preceding')
                    self._token('::')
            with self._option():
                with self._group():
                    self._token('ancestor-or-self')
                    self._token('::')
            self._error('no available options')

    @tatsumasu()
    def _abbrevReverseStep_(self):
        self._token('..')

    @tatsumasu()
    def _nodeTest_(self):
        with self._choice():
            with self._option():
                self._kindTest_()
            with self._option():
                self._nameTest_()
            self._error('no available options')

    @tatsumasu()
    def _nameTest_(self):
        with self._choice():
            with self._option():
                self._qname_()
            with self._option():
                self._wildcard_()
            self._error('no available options')

    @tatsumasu()
    def _wildcard_(self):
        with self._choice():
            with self._option():
                self._token('*')
            with self._option():
                with self._group():
                    self._ncname_()
                    self._token(':')
                    self._token('*')
            with self._option():
                with self._group():
                    self._token('*')
                    self._token(':')
                    self._ncname_()
            self._error('no available options')

    @tatsumasu()
    def _filterExpr_(self):
        self._primaryExpr_()
        self._predicateList_()

    @tatsumasu()
    def _predicateList_(self):

        def block0():
            self._predicate_()
        self._closure(block0)

    @tatsumasu()
    def _predicate_(self):
        self._token('[')
        self._expr_()
        self._token(']')

    @tatsumasu()
    def _primaryExpr_(self):
        with self._choice():
            with self._option():
                self._literal_()
            with self._option():
                self._varRef_()
            with self._option():
                self._parenthesizedExpr_()
            with self._option():
                self._contextItemExpr_()
            with self._option():
                self._functionCall_()
            self._error('no available options')

    @tatsumasu()
    def _literal_(self):
        with self._choice():
            with self._option():
                self._numericLiteral_()
            with self._option():
                self._stringLiteral_()
            self._error('no available options')

    @tatsumasu()
    def _numericLiteral_(self):
        with self._choice():
            with self._option():
                self._integerLiteral_()
            with self._option():
                self._decimalLiteral_()
            with self._option():
                self._doubleLiteral_()
            self._error('no available options')

    @tatsumasu()
    def _varRef_(self):
        self._token('$')
        self._varName_()

    @tatsumasu()
    def _varName_(self):
        self._qname_()

    @tatsumasu()
    def _parenthesizedExpr_(self):
        self._token('(')
        with self._optional():
            self._expr_()
        self._token(')')

    @tatsumasu()
    def _contextItemExpr_(self):
        self._token('.')

    @tatsumasu()
    def _functionCall_(self):
        self._qname_()
        self._token('(')
        with self._optional():
            self._exprSingle_()

            def block0():
                self._token(',')
                self._exprSingle_()
            self._closure(block0)
        self._token(')')

    @tatsumasu()
    def _singleType_(self):
        self._atomicType_()
        with self._optional():
            self._token('?')

    @tatsumasu()
    def _sequenceType_(self):
        with self._choice():
            with self._option():
                with self._group():
                    self._token('empty-sequence')
                    self._token('(')
                    self._token(')')
            with self._option():
                with self._group():
                    self._itemType_()
                    with self._optional():
                        self._occurrenceIndicator_()
            self._error('no available options')

    @tatsumasu()
    def _occurrenceIndicator_(self):
        with self._choice():
            with self._option():
                self._token('?')
            with self._option():
                self._token('*')
            with self._option():
                self._token('+')
            self._error('no available options')

    @tatsumasu()
    def _itemType_(self):
        with self._choice():
            with self._option():
                self._kindTest_()
            with self._option():
                with self._group():
                    self._token('item')
                    self._token('(')
                    self._token(')')
            with self._option():
                self._atomicType_()
            self._error('no available options')

    @tatsumasu()
    def _atomicType_(self):
        self._qname_()

    @tatsumasu()
    def _kindTest_(self):
        with self._choice():
            with self._option():
                self._documentTest_()
            with self._option():
                self._elementTest_()
            with self._option():
                self._attributeTest_()
            with self._option():
                self._schemaElementTest_()
            with self._option():
                self._schemaAttributeTest_()
            with self._option():
                self._pITest_()
            with self._option():
                self._commentTest_()
            with self._option():
                self._textTest_()
            with self._option():
                self._anyKindTest_()
            self._error('no available options')

    @tatsumasu()
    def _anyKindTest_(self):
        self._token('node')
        self._token('(')
        self._token(')')

    @tatsumasu()
    def _documentTest_(self):
        self._token('document-node')
        self._token('(')
        with self._optional():
            with self._choice():
                with self._option():
                    self._elementTest_()
                with self._option():
                    self._schemaElementTest_()
                self._error('no available options')
        self._token(')')

    @tatsumasu()
    def _textTest_(self):
        self._token('text')
        self._token('(')
        self._token(')')

    @tatsumasu()
    def _commentTest_(self):
        self._token('comment')
        self._token('(')
        self._token(')')

    @tatsumasu()
    def _pITest_(self):
        self._token('processing-instruction')
        self._token('(')
        with self._optional():
            with self._choice():
                with self._option():
                    self._ncname_()
                with self._option():
                    self._stringLiteral_()
                self._error('no available options')
        self._token(')')

    @tatsumasu()
    def _attributeTest_(self):
        self._token('attribute')
        self._token('(')
        with self._optional():
            self._attribNameOrWildcard_()
            with self._optional():
                self._token(',')
                self._typeName_()
        self._token(')')

    @tatsumasu()
    def _attribNameOrWildcard_(self):
        with self._choice():
            with self._option():
                self._attributeName_()
            with self._option():
                self._token('*')
            self._error('no available options')

    @tatsumasu()
    def _schemaAttributeTest_(self):
        self._token('schema-attribute')
        self._token('(')
        self._attributeDeclaration_()
        self._token(')')

    @tatsumasu()
    def _attributeDeclaration_(self):
        self._attributeName_()

    @tatsumasu()
    def _elementTest_(self):
        self._token('element')
        self._token('(')
        with self._optional():
            self._elementNameOrWildcard_()
            with self._optional():
                self._token(',')
                self._typeName_()
                with self._optional():
                    self._token('?')
        self._token(')')

    @tatsumasu()
    def _elementNameOrWildcard_(self):
        with self._choice():
            with self._option():
                self._elementName_()
            with self._option():
                self._token('*')
            self._error('no available options')

    @tatsumasu()
    def _schemaElementTest_(self):
        self._token('schema-element')
        self._token('(')
        self._elementDeclaration_()
        self._token(')')

    @tatsumasu()
    def _elementDeclaration_(self):
        self._elementName_()

    @tatsumasu()
    def _attributeName_(self):
        self._qname_()

    @tatsumasu()
    def _elementName_(self):
        self._qname_()

    @tatsumasu()
    def _typeName_(self):
        self._qname_()

    @tatsumasu()
    def _integerLiteral_(self):
        self._DIGITS_()

    @tatsumasu()
    def _decimalLiteral_(self):
        with self._choice():
            with self._option():
                with self._group():
                    self._token('.')
                    self._DIGITS_()
            with self._option():
                with self._group():
                    self._DIGITS_()
                    self._pattern(r'\.[0-9]*')
            self._error('no available options')

    @tatsumasu()
    def _doubleLiteral_(self):
        with self._group():
            with self._choice():
                with self._option():
                    with self._group():
                        self._token('.')
                        self._DIGITS_()
                with self._option():
                    with self._group():
                        self._DIGITS_()
                        self._pattern(r'(\.[0-9]*)?')
                self._error('no available options')
        self._pattern(r'[eE][+-]?')
        self._DIGITS_()

    @tatsumasu()
    def _stringLiteral_(self):
        with self._choice():
            with self._option():
                with self._group():
                    self._pattern(r'"')

                    def block0():
                        with self._choice():
                            with self._option():
                                self._ESCAPEQUOT_()
                            with self._option():
                                self._pattern(r'[^"]')
                            self._error('no available options')
                    self._closure(block0)
                    self._pattern(r'"')
            with self._option():
                with self._group():
                    self._pattern(r"'")

                    def block2():
                        with self._choice():
                            with self._option():
                                self._ESCAPEAPOS_()
                            with self._option():
                                self._pattern(r"[^']")
                            self._error('no available options')
                    self._closure(block2)
                    self._pattern(r"'")
            self._error('no available options')

    @tatsumasu()
    def _ESCAPEQUOT_(self):
        self._token('""')

    @tatsumasu()
    def _ESCAPEAPOS_(self):
        self._token("''")

    @tatsumasu()
    def _qname_(self):
        with self._choice():
            with self._option():
                with self._group():
                    self._NCNAME_FRAG_()
                    self._pattern(r':')
                    self._NCNAME_FRAG_()
            with self._option():
                self._NCNAME_FRAG_()
            self._error('no available options')

    @tatsumasu()
    def _ncname_(self):
        self._NCNAME_FRAG_()

    @tatsumasu()
    def _NCNAME_FRAG_(self):
        self._pattern(r'([A-Z]|_|[a-z]|[\u00C0-\u00D6]|[\u00D8-\u00F6]|[\u00F8-\u02FF]|[\u0370-\u037D]|[\u037F-\u1FFF]|[\u200C-\u200D]|[\u2070-\u218F]|[\u2C00-\u2FEF]|[\u3001-\uD7FF]|[\uF900-\uFDCF]|[\uFDF0-\uFFFD]|[\U00010000-\U000EFFFF])([A-Z]|_|[a-z]|[\u00C0-\u00D6]|[\u00D8-\u00F6]|[\u00F8-\u02FF]|[\u0370-\u037D]|[\u037F-\u1FFF]|[\u200C-\u200D]|[\u2070-\u218F]|[\u2C00-\u2FEF]|[\u3001-\uD7FF]|[\uF900-\uFDCF]|[\uFDF0-\uFFFD]|[\U00010000-\U000EFFFF]|-|\.|[0-9]|\u00B7|[\u0300-\u036F]|[\u203F-\u2040])*')

    @tatsumasu()
    def _DIGITS_(self):
        self._pattern(r'[0-9]+')


class XFSemantics(object):
    def module(self, ast):
        return ast

    def separator(self, ast):
        return ast

    def namespace_declaration(self, ast):
        return ast

    def default(self, ast):
        return ast

    def severity(self, ast):
        return ast

    def message_severity(self, ast):
        return ast

    def parameter(self, ast):
        return ast

    def filter_declaration(self, ast):
        return ast

    def filter(self, ast):
        return ast

    def concept_filter(self, ast):
        return ast

    def general_filter(self, ast):
        return ast

    def period_filter(self, ast):
        return ast

    def dimension_filter(self, ast):
        return ast

    def dimension_axis(self, ast):
        return ast

    def unit_filter(self, ast):
        return ast

    def entity_filter(self, ast):
        return ast

    def match_filter(self, ast):
        return ast

    def relative_filter(self, ast):
        return ast

    def tuple_filter(self, ast):
        return ast

    def value_filter(self, ast):
        return ast

    def boolean_filter(self, ast):
        return ast

    def aspect_cover_filter(self, ast):
        return ast

    def concept_relation_filter(self, ast):
        return ast

    def relation_axis(self, ast):
        return ast

    def declared_filter_reference(self, ast):
        return ast

    def fact_variable(self, ast):
        return ast

    def general_variable(self, ast):
        return ast

    def function_declaration(self, ast):
        return ast

    def assertion(self, ast):
        return ast

    def label(self, ast):
        return ast

    def message(self, ast):
        return ast

    def referenced_parameter(self, ast):
        return ast

    def precondition(self, ast):
        return ast

    def value_expression(self, ast):
        return ast

    def existence_expression(self, ast):
        return ast

    def enclosed_expression(self, ast):
        return ast

    def name(self, ast):
        return ast

    def language(self, ast):
        return ast

    def quoted_anyURI(self, ast):
        return ast

    def quoted_url(self, ast):
        return ast

    def localname(self, ast):
        return ast

    def variable_ref(self, ast):
        return ast

    def variable_name(self, ast):
        return ast

    def non_negative_integer(self, ast):
        return ast

    def date_time_constant(self, ast):
        return ast

    def date_constant(self, ast):
        return ast

    def quoted_string(self, ast):
        return ast

    def regexp_pattern(self, ast):
        return ast

    def xPath(self, ast):
        return ast

    def expr(self, ast):
        return ast

    def exprSingle(self, ast):
        return ast

    def forExpr(self, ast):
        return ast

    def simpleForClause(self, ast):
        return ast

    def quantifiedExpr(self, ast):
        return ast

    def ifExpr(self, ast):
        return ast

    def orExpr(self, ast):
        return ast

    def andExpr(self, ast):
        return ast

    def comparisonExpr(self, ast):
        return ast

    def rangeExpr(self, ast):
        return ast

    def additiveExpr(self, ast):
        return ast

    def multiplicativeExpr(self, ast):
        return ast

    def unionExpr(self, ast):
        return ast

    def intersectExceptExpr(self, ast):
        return ast

    def instanceofExpr(self, ast):
        return ast

    def treatExpr(self, ast):
        return ast

    def castableExpr(self, ast):
        return ast

    def castExpr(self, ast):
        return ast

    def unaryExpr(self, ast):
        return ast

    def valueExpr(self, ast):
        return ast

    def generalComp(self, ast):
        return ast

    def valueComp(self, ast):
        return ast

    def nodeComp(self, ast):
        return ast

    def pathExpr(self, ast):
        return ast

    def relativePathExpr(self, ast):
        return ast

    def stepExpr(self, ast):
        return ast

    def axisStep(self, ast):
        return ast

    def forwardStep(self, ast):
        return ast

    def forwardAxis(self, ast):
        return ast

    def abbrevForwardStep(self, ast):
        return ast

    def reverseStep(self, ast):
        return ast

    def reverseAxis(self, ast):
        return ast

    def abbrevReverseStep(self, ast):
        return ast

    def nodeTest(self, ast):
        return ast

    def nameTest(self, ast):
        return ast

    def wildcard(self, ast):
        return ast

    def filterExpr(self, ast):
        return ast

    def predicateList(self, ast):
        return ast

    def predicate(self, ast):
        return ast

    def primaryExpr(self, ast):
        return ast

    def literal(self, ast):
        return ast

    def numericLiteral(self, ast):
        return ast

    def varRef(self, ast):
        return ast

    def varName(self, ast):
        return ast

    def parenthesizedExpr(self, ast):
        return ast

    def contextItemExpr(self, ast):
        return ast

    def functionCall(self, ast):
        return ast

    def singleType(self, ast):
        return ast

    def sequenceType(self, ast):
        return ast

    def occurrenceIndicator(self, ast):
        return ast

    def itemType(self, ast):
        return ast

    def atomicType(self, ast):
        return ast

    def kindTest(self, ast):
        return ast

    def anyKindTest(self, ast):
        return ast

    def documentTest(self, ast):
        return ast

    def textTest(self, ast):
        return ast

    def commentTest(self, ast):
        return ast

    def pITest(self, ast):
        return ast

    def attributeTest(self, ast):
        return ast

    def attribNameOrWildcard(self, ast):
        return ast

    def schemaAttributeTest(self, ast):
        return ast

    def attributeDeclaration(self, ast):
        return ast

    def elementTest(self, ast):
        return ast

    def elementNameOrWildcard(self, ast):
        return ast

    def schemaElementTest(self, ast):
        return ast

    def elementDeclaration(self, ast):
        return ast

    def attributeName(self, ast):
        return ast

    def elementName(self, ast):
        return ast

    def typeName(self, ast):
        return ast

    def integerLiteral(self, ast):
        return ast

    def decimalLiteral(self, ast):
        return ast

    def doubleLiteral(self, ast):
        return ast

    def stringLiteral(self, ast):
        return ast

    def ESCAPEQUOT(self, ast):
        return ast

    def ESCAPEAPOS(self, ast):
        return ast

    def qname(self, ast):
        return ast

    def ncname(self, ast):
        return ast

    def NCNAME_FRAG(self, ast):
        return ast

    def DIGITS(self, ast):
        return ast


def main(filename, start=None, **kwargs):
    if start is None:
        start = 'module'
    if not filename or filename == '-':
        text = sys.stdin.read()
    else:
        with open(filename) as f:
            text = f.read()
    parser = XFParser()
    return parser.parse(text, rule_name=start, filename=filename, **kwargs)


if __name__ == '__main__':
    import json
    from tatsu.util import asjson

    ast = generic_main(main, XFParser, name='XF')
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(asjson(ast), indent=2))
    print()
